﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.Linq;
using System.Windows;
using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.WpfExplorer;

namespace ThinkGeo.MapSuite.WpfDesktopEdition.Extension
{
    public class MeasureTrackInteractiveOverlay : TrackInteractiveOverlay
    {
        private static readonly Feature defaultValueOfFeature = default(Feature);
        private static readonly string valueStyleColumnName = "Value_Style_Key";

        private GeoBrush fillColor;
        private GeoBrush outlineColor;
        private float outlineThickness;

        private IconTextStyle textStyle;

        //column name should be less than 10 chars. because dbf file requires so
        private static readonly string MeasureResultColumnName = "Result";
        private static readonly string MemoColumnName = "Memo";

        public static string AreaColumnName { get { return "Area"; } }
        public static string DistanceColumnName { get { return "Distance"; } }
        public static string UnitColumnName { get { return "Unit"; } }

        private string valueStyleColumnValue = "default";

        private Stack<Feature> historyFeatures;

        public MeasureTrackInteractiveOverlay()
        {
            RenderMode = RenderMode.GdiPlus;

            TrackShapeLayer.Open();

            TrackShapeLayer.Columns.Add(new FeatureSourceColumn(MeasureResultColumnName));
            TrackShapeLayer.Columns.Add(new FeatureSourceColumn(MemoColumnName));
            TrackShapeLayer.Columns.Add(new FeatureSourceColumn(valueStyleColumnName));

            TrackShapeLayer.Columns.Add(new FeatureSourceColumn(AreaColumnName));
            TrackShapeLayer.Columns.Add(new FeatureSourceColumn(DistanceColumnName));
            TrackShapeLayer.Columns.Add(new FeatureSourceColumn(UnitColumnName));

            TrackShapeLayer.Close();

            outlineColor = new GeoSolidBrush(new GeoColor(255, 255, 0, 0));
            //fillColor = new GeoSolidBrush(new GeoColor(128, 255, 255, 255));
            fillColor = new GeoSolidBrush(GeoColor.SimpleColors.Transparent);

            outlineThickness = 1;

            textStyle = new IconTextStyle()
            {
                Name = "TextStyle",
                TextColumnName = MeasureResultColumnName,
                Font = new GeoFont("Arial", 12),
                TextSolidBrush = new GeoSolidBrush(GeoColor.SimpleColors.Black),
                BestPlacement = false,
                DuplicateRule = LabelDuplicateRule.UnlimitedDuplicateLabels,
                OverlappingRule = LabelOverlappingRule.AllowOverlapping,
                ForceLineCarriage = true,
                SuppressPartialLabels = false,
                SplineType = SplineType.StandardSplining,
                TextLineSegmentRatio = 10000,
                AllowLineCarriage = true,
            };

            InitValueStyle();
            historyFeatures = new Stack<Feature>();
        }

        public bool CanRollback
        {
            get
            {
                return TrackShapeLayer.InternalFeatures.Count > 0;
            }
        }

        public Stack<Feature> HistoryFeatures
        {
            get
            {
                return historyFeatures;
            }
        }

        public bool CanForward
        {
            get
            {
                return historyFeatures.Count > 0;
            }
        }

        public override bool IsEmpty
        {
            get
            {
                return false;
            }
        }

        public GeoBrush FillColor
        {
            get { return fillColor; }
            set
            {
                if (value != fillColor)
                {
                    fillColor = value;
                    UpdateValueStyle();
                }
            }
        }

        public GeoBrush OutlineColor
        {
            get { return outlineColor; }
            set
            {
                if (value != outlineColor)
                {
                    outlineColor = value;
                    UpdateValueStyle();
                }
            }
        }

        public float OutlineThickness
        {
            get { return outlineThickness; }
            set
            {
                if (value != outlineThickness)
                {
                    outlineThickness = value;
                    UpdateValueStyle();
                }
            }
        }

        public IconTextStyle TextStyle
        {
            get { return textStyle; }
            set
            {
                textStyle = value;
                textStyle.SplineType = SplineType.StandardSplining;
                textStyle.TextColumnName = MeasureResultColumnName;
                UpdateValueStyle();
            }
        }

        public void Rollback()
        {
            if (CanRollback)
            {
                Feature feature = TrackShapeLayer.InternalFeatures[TrackShapeLayer.InternalFeatures.Count - 1];
                TrackShapeLayer.InternalFeatures.Remove(feature);
                historyFeatures.Push(feature);
                this.Refresh();
            }
        }

        public void Forward()
        {
            if (CanForward)
            {
                Feature feature = historyFeatures.Pop();
                TrackShapeLayer.InternalFeatures.Add(feature);
                this.Refresh();
            }
        }

        /// <summary>
        /// when this property is set to be true and trackmode is none, we can single click to select a feature
        /// that is actually in other layers
        /// </summary>
        public bool IsInSelectMode { get; set; }

        public DistanceUnit DistanceUnit
        {
            get;
            set;
        }

        public AreaUnit AreaUnit
        {
            get;
            set;
        }

        protected override InteractiveResult MouseMoveCore(InteractionArguments interactionArguments)
        {
            InteractiveResult interactiveResult = base.MouseMoveCore(interactionArguments);
            MeasureFeature();

            return interactiveResult;
        }

        protected override InteractiveResult MouseClickCore(InteractionArguments interactionArguments)
        {
            if (TrackMode == WpfDesktopEdition.TrackMode.Custom && IsInSelectMode)
            {
                Feature foundFeature = FindFeatureByPoint(interactionArguments.WorldX, interactionArguments.WorldY);
                if (!IsFoundFeatureMeasured(foundFeature))
                {
                    if (foundFeature != defaultValueOfFeature)
                    {
                        TrackShapeLayer.InternalFeatures.Add(foundFeature);
                        historyFeatures.Clear();
                        MeasureFeature();
                        Refresh();
                    }
                    else
                    {
                        MessageBox.Show("No features found.", "Info", MessageBoxButton.OK, MessageBoxImage.Information);
                    }
                }
            }
            return base.MouseClickCore(interactionArguments);
        }

        protected override void OnTrackEnded(TrackEndedTrackInteractiveOverlayEventArgs e)
        {
            base.OnTrackEnded(e);
            MeasureFeature();
            historyFeatures.Clear();
        }

        protected override byte[] SaveStateCore()
        {
            Queue<object> stateQueue = new Queue<object>();
            stateQueue.Enqueue(base.SaveStateCore());
            stateQueue.Enqueue(FillColor);
            stateQueue.Enqueue(OutlineColor);
            return SerializeHelper.Serialize(stateQueue);
        }

        protected override void LoadStateCore(byte[] state)
        {
            Queue<object> stateQueue = SerializeHelper.Deserialize<Queue<object>>(state);
            base.LoadStateCore((byte[])stateQueue.Dequeue());
            FillColor = (GeoBrush)stateQueue.Dequeue();
            OutlineColor = (GeoBrush)stateQueue.Dequeue();
        }

        private void MeasureFeature()
        {
            if (TrackShapeLayer.InternalFeatures.Count > 0)
            {
                Feature feature = TrackShapeLayer.InternalFeatures[TrackShapeLayer.InternalFeatures.Count - 1];
                if (!feature.ColumnValues.ContainsKey(MeasureResultColumnName))
                {
                    //string textValue = string.Empty;
                    switch (TrackMode)
                    {
                        case TrackMode.Rectangle:
                        case TrackMode.Square:
                        case TrackMode.Ellipse:
                        case TrackMode.Circle:
                        case TrackMode.Polygon:
                            SetTextForArea(feature);
                            break;
                        case TrackMode.Line:
                            SetTextForLine(feature);
                            break;
                        case TrackMode.Custom:
                            BaseShape baseShape = feature.GetShape();
                            if (baseShape is AreaBaseShape)
                            {
                                SetTextForArea(feature);
                            }
                            else if (baseShape is LineBaseShape)
                            {
                                SetTextForLine(feature);
                            }
                            break;
                        case TrackMode.Point:
                            break;
                    }
                    feature.ColumnValues[valueStyleColumnName] = valueStyleColumnValue;
                    //feature.ColumnValues[MeasureResultColumnName] = textValue;

                    Feature greatCircleFeature = GetGreatCircle(feature);
                    if (greatCircleFeature != defaultValueOfFeature)
                    {
                        TrackShapeLayer.InternalFeatures[TrackShapeLayer.InternalFeatures.Count - 1] = greatCircleFeature;
                    }
                }
                //this.Refresh();
            }
        }

        private void SetTextForLine(Feature feature)
        {
            double length = 0;
            string textValue = string.Empty;
            try
            {
                length = ((LineBaseShape)feature.GetShape()).GetLength(MapArguments.MapUnit, DistanceUnit);
                textValue = string.Format(CultureInfo.InvariantCulture, "{0:N3} {1}", length, GetAbbreviateDistanceUnit(DistanceUnit));
            }
            catch
            {
                textValue = "Error";
            }
            finally
            {
                feature.ColumnValues[DistanceColumnName] = length.ToString();
                feature.ColumnValues[UnitColumnName] = DistanceUnit.ToString();
                feature.ColumnValues[MeasureResultColumnName] = textValue;
            }
        }

        //from http://englishplus.com/grammar/00000058.htm
        private static string GetAbbreviateDistanceUnit(DistanceUnit distanceUnit)
        {
            switch (distanceUnit)
            {
                case DistanceUnit.Meter:
                    return "m";
                case DistanceUnit.Feet:
                    return "ft.";
                case DistanceUnit.Kilometer:
                    return "km";
                case DistanceUnit.Mile:
                    return "mi.";
                case DistanceUnit.UsSurveyFeet:
                    return "us-ft.";
                case DistanceUnit.Yard:
                    return "yd.";
                default:
                    return "unkown.";
            }
        }

        private void SetTextForArea(Feature feature)
        {
            double area = ((AreaBaseShape)feature.GetShape()).GetArea(MapArguments.MapUnit, AreaUnit);
            string textValue = string.Format("{0:N3} {1}", area, GetAbbreviateAreaUnit(AreaUnit));

            feature.ColumnValues[AreaColumnName] = area.ToString();
            feature.ColumnValues[UnitColumnName] = AreaUnit.ToString();
            feature.ColumnValues[MeasureResultColumnName] = textValue;
        }

        private static string GetAbbreviateAreaUnit(AreaUnit areaUnit)
        {
            switch (areaUnit)
            {
                case AreaUnit.SquareMeters:
                    return "m2";
                case AreaUnit.Acres:
                    return "ac.";
                case AreaUnit.Hectares:
                    return "ha";
                case AreaUnit.SquareFeet:
                    return "sq. ft.";
                case AreaUnit.SquareKilometers:
                    return "sq. km";
                case AreaUnit.SquareMiles:
                    return "sq. mi.";
                case AreaUnit.SquareUsSurveyFeet:
                    return "sq. us-ft.";
                case AreaUnit.SquareYards:
                    return "sq. yd.";
                default:
                    return "unknown.";
            }
        }

        private Feature GetGreatCircle(Feature feature)
        {
            LineShape line = feature.GetShape() as LineShape;
            if (line != null)
            {
                MultilineShape greatCircle = new MultilineShape();
                for (int i = 0; i < line.Vertices.Count; i++)
                {
                    if (i != line.Vertices.Count - 1)
                    {
                        PointShape point = new PointShape(line.Vertices[i]);
                        MultilineShape multiline = point.GreatCircle(new PointShape(line.Vertices[i + 1]));
                        if (multiline != null)
                        {
                            multiline.Lines.ForEach(l =>
                            {
                                greatCircle.Lines.Add(l);
                            });
                        }
                    }
                }

                if (greatCircle.Lines.Count > 0)
                {
                    Feature wrapperFeature = new Feature(greatCircle);
                    feature.ColumnValues.ForEach(columnValue =>
                    {
                        //Console.WriteLine(columnValue.Key + "   " + columnValue.Value);
                        wrapperFeature.ColumnValues.Add(columnValue.Key, columnValue.Value);
                    });

                    return wrapperFeature;
                }
            }

            return defaultValueOfFeature;
        }

        private void UpdateValueStyle()
        {
            RefreshValueStyleColumnValue();
            ValueStyle trackLayerStyle = TrackShapeLayer.ZoomLevelSet.ZoomLevel01.CustomStyles[0] as ValueStyle;
            if (trackLayerStyle != null)
            {
                trackLayerStyle.ValueItems.Add(CreatValueItem());
            }
        }

        private ValueItem CreatValueItem()
        {
            ValueItem result = new ValueItem { Value = valueStyleColumnValue };

            result.DefaultAreaStyle = new AreaStyle();
            result.DefaultAreaStyle.Advanced.FillCustomBrush = FillColor;
            result.DefaultAreaStyle.OutlinePen.Brush = OutlineColor;
            result.DefaultAreaStyle.OutlinePen.Width = OutlineThickness;

            result.DefaultLineStyle = new LineStyle();
            result.DefaultLineStyle.InnerPen.Brush = FillColor;
            result.DefaultLineStyle.OuterPen.Brush = OutlineColor;
            result.DefaultLineStyle.OuterPen.Width = OutlineThickness;

            result.DefaultPointStyle = new PointStyle();
            result.DefaultPointStyle.Advanced.CustomBrush = FillColor;
            result.DefaultPointStyle.SymbolPen.Brush = OutlineColor;
            result.DefaultPointStyle.SymbolPen.Width = OutlineThickness;

            //result.DefaultTextStyle = new TextStyle(this.TextStyle.TextColumnName, this.TextStyle.Font, this.TextStyle.TextSolidBrush)
            //{
            //    Name = this.TextStyle.Name,
            //    BestPlacement = false,
            //    DuplicateRule = LabelDuplicateRule.UnlimitedDuplicateLabels,
            //    OverlappingRule = LabelOverlappingRule.AllowOverlapping,
            //    ForceLineCarriage = true,
            //    AllowLineCarriage = true,
            //    SuppressPartialLabels = false,
            //    SplineType = SplineType.StandardSplining,
            //    TextLineSegmentRatio = 10000,
            //};

            TextStyle.BestPlacement = false;
            TextStyle.DuplicateRule = LabelDuplicateRule.UnlimitedDuplicateLabels;
            TextStyle.OverlappingRule = LabelOverlappingRule.AllowOverlapping;
            TextStyle.ForceLineCarriage = true;
            TextStyle.AllowLineCarriage = true;
            TextStyle.SuppressPartialLabels = false;
            TextStyle.SplineType = SplineType.StandardSplining;
            TextStyle.TextLineSegmentRatio = 10000;

            result.DefaultTextStyle = TextStyle;

            //result.CustomStyles.Add(result.DefaultAreaStyle);
            //result.CustomStyles.Add(result.DefaultLineStyle);
            //result.CustomStyles.Add(result.DefaultPointStyle);
            //result.CustomStyles.Add(result.DefaultTextStyle);

            //result.DefaultAreaStyle = null;
            //result.DefaultLineStyle = null;
            //result.DefaultPointStyle = null;
            //result.DefaultTextStyle = null;

            return result;
        }

        private void RefreshValueStyleColumnValue()
        {
            valueStyleColumnValue = DateTime.Now.GetHashCode().ToString();
        }

        private void InitValueStyle()
        {
            ValueStyle trackLayerStyle = new ValueStyle { ColumnName = valueStyleColumnName };

            TrackShapeLayer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
            TrackShapeLayer.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(trackLayerStyle);

            TrackShapeLayer.ZoomLevelSet.ZoomLevel01.DefaultAreaStyle = null;
            TrackShapeLayer.ZoomLevelSet.ZoomLevel01.DefaultLineStyle = null;
            TrackShapeLayer.ZoomLevelSet.ZoomLevel01.DefaultPointStyle = null;
            TrackShapeLayer.ZoomLevelSet.ZoomLevel01.DefaultTextStyle = null;

            trackLayerStyle.ValueItems.Add(CreatValueItem());
        }

        private bool IsFoundFeatureMeasured(Feature foundFeature)
        {
            bool isMeasured = false;

            if (foundFeature != defaultValueOfFeature)
            {
                foreach (Feature feature in TrackShapeLayer.InternalFeatures)
                {
                    if (feature.GetWellKnownText() == foundFeature.GetWellKnownText())
                    {
                        isMeasured = true;
                        break;
                    }
                }
            }

            return isMeasured;
        }

        private Feature FindFeatureByPoint(double x, double y)
        {
            Feature result = defaultValueOfFeature;

            var allFeatureLayers = Globals.ActiveMap.Overlays
                                   .Reverse()
                                   .OfType<LayerOverlay>()
                                   .Where(layerOverlay => layerOverlay.IsVisible)
                                   .SelectMany(layerOverlay => layerOverlay.Layers
                                                                .Reverse()
                                                                .OfType<FeatureLayer>()
                                                                .Where(layer => layer.IsVisible));

            GeographyUnit currentUnit = GeographyUnit.DecimalDegree;
            string currentProj4 = ((ExtendedWpfMap)Globals.ActiveMap).DisplayProjectionParameters;
            if (!string.IsNullOrEmpty(currentProj4))
            {
                currentUnit = MapReprojector.GetGeographyUnit(currentProj4);
            }

            foreach (var featureLayer in allFeatureLayers)
            {
                featureLayer.Open();
                var featuresInDistance = AscendingSearch(x, y, currentUnit, featureLayer, 1, 20);
                featureLayer.Close();

                if (featuresInDistance != null && featuresInDistance.Count > 0)
                {
                    result = featuresInDistance[0];
                    break;
                }
            }

            return result;
        }

        private static Collection<Feature> AscendingSearch(double x, double y, GeographyUnit currentUnit, FeatureLayer featureLayer, double lowerScale, double upperScale)
        {
            Collection<Feature> results = null;

            for (double i = lowerScale; i < upperScale; i++)
            {
                var featuresInDistance = featureLayer.FeatureSource
                                                     .GetFeaturesWithinDistanceOf(new Feature(x, y),
                                                      currentUnit,
                                                      DistanceUnit.Meter, lowerScale,
                                                      ReturningColumnsType.NoColumns);

                if (featuresInDistance.Count > 0)
                {
                    results = featuresInDistance;
                    break;
                }
            }

            return results;
        }
    }
}
